/*
  * Copyright (c) 2001-2003 Shiman Associates Inc. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/* $Id $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/audio.h>
#include <math.h>
#include "mas/mas_dpi.h"
#include "anx_internal.h"
#include "aixppc.h"
#include "aixppc_profile.h"
#include "aixppc_defines.h"



/*************************************************************************
 * LOCAL PROTOTYPES
 *************************************************************************/
static int32 get_device_caps( struct anx_state* state );
static int32 aix_set_fragment_stuff( struct anx_state* state );
static int32 aix_set_rec_fragment_stuff( struct anx_state* state );
static int32 aix_add_mix_channel(struct anx_state* state, char* name, int32 portnum, int aix_ch_num);

static int aix_open_audio_device_playrec( struct anx_state* state );
static int aix_open_audio_device_mix( struct anx_state* state );
static int aix_open_audio_device_mstvol( struct anx_state* state );

static int tabulated_gain( int aix_channel, float db );
static float inverse_tabulated_gain( int aix_channel, int gain );

static void initialize_audio_change( audio_change *ac );

static int32 unmute_mic( struct anx_state* state );


/* TODO: I wish there was a way to detect capabilities.... well we
   could attempt to map back the stuff returned by AUDIO_GET_CHANNELS */

/* in sys/audio.h, input and output channels of different names may
   have the same values... */
static int sensible_channel_map[][2] =
{
    /* aix channel, is_input */
    { MASTER_VOLUME_CHANNEL, 0 },
    { AUDIO_OUTPUT_1       , 0 },
    { AUDIO_HIGH_GAIN_MIKE , 1 },
    { -1, -1 }
};

static int is_input_channel( int ch )
{
    return sensible_channel_map[ch][1];
}
static int aix_channel( int ch )
{
    return sensible_channel_map[ch][0];
}


/* BEGIN PLATFORM DEPENDENT ACTIONS ***********************************/



int32
pdanx_init_library( void )
{
    /* nothing to do */
    return 0;
}

int32
pdanx_exit_library( void )
{
    /* nothing to do */
    return 0;
}

int32
pdanx_init_instance( struct anx_state* state, void* predicate )
{
    int32                    err;

    /* null out the platform-dependent state */
    memset( &(state->pdstate), 0, sizeof state->pdstate );

    /* special cases */
    state->pdstate.play_fd    = -1;
    state->pdstate.rec_fd     = -1;
    state->pdstate.mix_fd     = -1;
    state->pdstate.mstvol_fd  = -1;
    state->no_loop_record     = TRUE;

    state->pdstate.internal_speaker_state = 1;

    strncpy( state->pdstate.devfname, DEFAULT_AUDIO_DEVICE, 128 );
    
        
    /* open the sound device */
    err = pdanx_open_resource( state );
    if ( err < 0 )
        return err;
    
    /* get device capabilities */
    err = get_device_caps( state );
    if ( err < 0 )
        return err;

    /* make the buffer */
    err = anx_make_buffer( &state->play_buffer, PLAY_BUFFER_SIZE, 0 );
    if ( err < 0 )
    {
        masc_exiting_log_level();
        return err;
    }
    
    return 0;
}

/* Open the audio device
   TODO: don't assume /dev/paud0 */
int32
pdanx_open_resource( struct anx_state* state )
{
    int   holder;
    int32 err;
    audio_init ai;

    
    /* wait... is it already open?! bail if it is. */
    if ( state->pdstate.play_fd >= 0 )
        return 0;
    
    /* open audio device channel */
    err = aix_open_audio_device_playrec( state );
    if( err<0 ) return err;
    if ( state->pdstate.play_fd < 0 )
        return mas_error( MERR_FILE_CANNOT_OPEN );    
    
    
/*     /\* open mixer device channel *\/ */
/*     err = aix_open_audio_device_mix( state ); */
/*     if( err<0 ) return err; */
    
/*     if ( state->pdstate.mix_fd < 0 ) */
/*         return mas_error( MERR_FILE_CANNOT_OPEN ); */
    

/*     /\* open master volume device channel *\/ */
/*     err = aix_open_audio_device_mstvol( state ); */
/*     if( err<0 ) return err; */
/*     if ( state->pdstate.mstvol_fd < 0 ) */
/*         return mas_error( MERR_FILE_CANNOT_OPEN ); */
    
    return 0;
}


/* Close the audio device */
int32
pdanx_close_resource( struct anx_state* state )
{
    int32 err;

    err = ioctl( state->pdstate.play_fd, AUDIO_WAIT, NULL );
    if( err<0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "error %d draining kernel playback audio buffers", err );
    }

#ifdef DEBUG
    masc_log_message( MAS_VERBLVL_DEBUG, "closing resource" );
#endif

    close( state->pdstate.play_fd );
    close( state->pdstate.rec_fd );
/*     close( state->pdstate.mix_fd ); */
/*     close( state->pdstate.mstvol_fd ); */

    state->pdstate.play_fd    = -1;
    state->pdstate.rec_fd     = -1;
    state->pdstate.mstvol_fd  = -1;
    state->pdstate.rec_fd     = -1;

    state->pdstate.play_configured = FALSE;
    state->pdstate.rec_configured  = FALSE;

    return 0;
}

int32
pdanx_exit_instance( struct anx_state* state, void* predicate )
{
    pdanx_close_resource( state );
    return 0;
}


/* Handle state transitions.  Open and re-configure the device if
 * transitioning from inactive/inactive_pending to
 * active/active_pending. */
int32
pdanx_change_res_state( struct anx_state *state, enum res_state res_state )
{
    int32 err;
    
    /* act based on prior state */
    if ( state->res_state == ACTIVE || state->res_state == ACTIVE_PENDING )
    {
        /* Switching between active states requires no action. */
        /* Switching to either inactive state requires closing the
           device. */
        if ( res_state == INACTIVE || res_state == INACTIVE_PENDING )
            pdanx_close_resource( state );
    }
    else /* inactive or inactive_pending */
    {
        /* switching between inactive states requires no action. */
        /* Switching to either active state: */

        if ( res_state == ACTIVE || res_state == ACTIVE_PENDING )
        {
            err = pdanx_open_resource( state );
            if ( err < 0 )
            {
                masc_log_message( MAS_VERBLVL_DEBUG, "anx: error opening resource" );
                return err;
            }
            
            /* if either of the ports are active, we have to configure
               the driver */
            if( state->source_active )
            {
                pdanx_configure_resource( state, state->audio_source );
                pdanx_record_start( state ); /* for the AUDIO_START ioctl */
            }

            if( state->sink_active )
            {
                pdanx_configure_resource( state, state->audio_sink );
                pdanx_playback_start( state ); /* for the AUDIO_START ioctl */
            }
        }
    }
    
    return 0;
}


/* Given the formats specified in the anx_state: play_format,
 * play_sample_rate, play_resolution, play_channels, and rec_format,
 * rec_sample_rate, rec_resolution, and rec_channels, configure the
 * device driver.  This same function serves both playback and
 * recording. */
int32
pdanx_configure_resource( struct anx_state *state, int32 portnum )
{
    int32         err;
    int           aix_format = -1;
    int           aix_flags;
    audio_init    ai;
    audio_control actl;
    audio_change  achg;
    

#ifdef MAS_LITTLE_ENDIAN
    aix_flags = AUDIO_FIXED;
#else
    aix_flags = AUDIO_FIXED | AUDIO_BIG_ENDIAN;
#endif
    
    switch ( state->play_format )
    {
    case MAS_ULAW_FMT:
        aix_format = AUDIO_MU_LAW;
        break;
    case MAS_ALAW_FMT:
        aix_format = AUDIO_A_LAW;
        break;
    case MAS_LINEAR_FMT:
        aix_format = AUDIO_PCM;
        aix_flags |= AUDIO_TWOS_COMPLEMENT;
        break;
    case MAS_ULINEAR_FMT:
        aix_format = AUDIO_PCM;
        /* ?! what's the appropriate flag in sys/audio.h? maybe
         * nothing is fine. */
        aix_flags = aix_flags;
        break;
    default:
        return mas_error(MERR_INVALID);
    }

    
    if ( (portnum==state->audio_sink) && (!state->pdstate.play_configured) )
    {
        /* we're configuring the playback interface */

        /* INITIALIZE */        
        ai.srate               = state->play_sample_rate; 
        ai.bits_per_sample     = state->play_resolution; 
        ai.bsize               = DEFAULT_PLAY_BLOCK_SIZE; /* see aixppc.h */
        ai.mode                = aix_format;
        ai.channels            = state->play_channels;
        ai.flags               = aix_flags;
        ai.operation           = AUDIO_PLAY;
        
        ai.position_resolution = 1;
        ai.rc                  = AUDIO_IGNORE;
        ai.slot_number         = AUDIO_IGNORE;
        ai.device_id           = AUDIO_IGNORE;
        *ai.loadpath           = 0;
        ai.reserved            = 0;
        
        
        err = ioctl( state->pdstate.play_fd, AUDIO_INIT, &ai );
        
        if( err < 0 )
        {
            masc_log_message( MAS_VERBLVL_DEBUG, "AUDIO_INIT (playback) ioctl returned %d: %s, errno==%d\n", err, error_symbol(ai.rc), errno );
            return mas_error( MERR_IO );
        }

        /* TODO:
           we may not get what we asked for :(  --need to be able to recover from
           that! */
        if( (ai.flags & AUDIO_BESTFIT_PROVIDED) != 0 )
        {
            masc_log_message( MAS_VERBLVL_WARNING, "anx: WARNING: didn't get the audio adapter settings we asked for:" );
            print_init( &ai );
        }   
#ifdef DEBUG
        else
        {
            print_init( &ai );
        }  
#endif

        err = aix_set_fragment_stuff( state );  
        if( err<0 )
        {
            return err;
        }
        else
        {
            masc_log_message( MAS_VERBLVL_DEBUG, "Allocated %d fragments of %d bytes each.", state->pdstate.fragments, state->pdstate.fragsize);
        }
        
        initialize_audio_change( &achg );
        achg.output  = DEFAULT_PLAY_CHANNEL;
        achg.volume  = 0x7fffffff / 2;
        achg.balance = 0x3FFFFFFF;
        
        actl.ioctl_request = AUDIO_CHANGE;
        actl.request_info = &achg;
        
        err = ioctl( state->pdstate.play_fd, AUDIO_CONTROL, &actl );
        if ( err<0 ) 
        {
            masc_log_message( MAS_VERBLVL_DEBUG, " AUDIO_CONTROL ioctl fro volume setting returned %d, errno '%s", err, strerror(errno) );
            return mas_error(MERR_INVALID);
        }

        /* port is now configured */
        state->pdstate.play_configured = TRUE;

        pdanx_get_mixer_volume( state, state->dac_ch ); /* now it's safe to do this. */
    }
    
    
    if ( portnum == state->audio_sink )
    {
        /* reset the playback buffer */
        state->play_buffer->fill_line = state->buftime_mt * state->play_bpstc;
        anx_reset_buffer( state->play_buffer );
        state->play_buffer->filling = TRUE;
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: playback filling %.1fms buffer", 1000.0 * (float) state->play_buffer->fill_line / (state->play_sample_rate * state->play_bpstc ) );
    }
    else
    {
        if( (portnum==state->audio_source) && (!state->pdstate.rec_configured) )
        {
            if( state->is_full_duplex )
            {
                /* INITIALIZE */
                ai.srate               = state->rec_sample_rate;
                ai.bits_per_sample     = state->rec_resolution; 
                ai.bsize               = MAS_ANX_SEGLEN;  /* TODO: tweak */
                ai.mode                = aix_format;
                ai.channels            = state->rec_channels; 
                ai.flags               = aix_flags;
                ai.operation           = AUDIO_RECORD;
                
                ai.position_resolution = 1;
                ai.rc                  = AUDIO_IGNORE;
                ai.slot_number         = AUDIO_IGNORE;
                ai.device_id           = AUDIO_IGNORE;
                *ai.loadpath           = 0;
                ai.reserved            = 0;
                
        
                err = ioctl( state->pdstate.rec_fd, AUDIO_INIT, &ai );
                if( err < 0 )
                {
                    masc_log_message( MAS_VERBLVL_DEBUG, "AUDIO_INIT (record) ioctl returned %d: %s, errno==%d\n", err, error_symbol(ai.rc), errno );
                    return mas_error( MERR_IO );
                }

                if( (ai.flags & AUDIO_BESTFIT_PROVIDED) != 0 )
                {
                    masc_log_message( MAS_VERBLVL_WARNING, "anx: WARNING: didn't get the audio adapter settings we asked for:" );
                    print_init( &ai );
                }   
#ifdef DEBUG
                else
                {
                    print_init( &ai );
                }
#endif

                err = aix_set_rec_fragment_stuff( state ); /* ???!
                                                              again? */
                if( err<0 )
                {
                    return err;
                }
                
                
                /* set the recording source */
                state->pdstate.rec_source = DEFAULT_REC_SOURCE;
                initialize_audio_change( &achg );
                achg.input                = DEFAULT_REC_SOURCE;
                
                achg.volume  = 0x7fffffff;
                achg.balance = 0x3FFFFFFF;
                
                achg.monitor  = AUDIO_MONITOR_OFF;
                
                actl.ioctl_request = AUDIO_CHANGE;
                actl.request_info = &achg;
                
                err = ioctl( state->pdstate.rec_fd, AUDIO_CONTROL, &actl );
                if ( err<0 )
                {
                    masc_log_message( MAS_VERBLVL_DEBUG, "rec AUDIO_CONTROL ioctl returned %d, errno '%s", err, strerror(errno) );
                    return mas_error(MERR_INVALID);
                }
                
#ifdef DEBUG
                {
                    audio_buffer ab;
                    
                    memset( &ab, 0, sizeof ab );            
                    err = ioctl( state->pdstate.rec_fd, AUDIO_BUFFER, &ab );
                    if( err < 0 )
                    {
                        masc_log_message( MAS_VERBLVL_DEBUG, "AUDIO_BUFFER rec ioctl returned %d. errno==%d '%s'", err, errno, strerror(errno) );
                        return mas_error(MERR_INVALID);
                    }
                    print_buffer( &ab );
                }
#endif

            }

            /* port is now configured */
            state->pdstate.rec_configured = TRUE;
            pdanx_get_mixer_volume( state, state->mic_ch ); /* now it's safe to do this. */
        }
        else
            return mas_error( MERR_INVALID );
    }
    return 0;
}
    

int32
pdanx_configure_port( struct anx_state* state, int32 portnum, struct mas_data_characteristic* dc )
{
    return pdanx_configure_resource( state, portnum );
}

int32
pdanx_disconnect_port( struct anx_state* state, int32 portnum )
{
    return 0;
}

int32
pdanx_playback_start( struct anx_state* state )
{
    int32 err;
    audio_control actl;
    audio_change achg;
                
    actl.ioctl_request = AUDIO_START;
    actl.position = 0;
    actl.request_info = NULL;

#ifdef DEBUG
    masc_log_message( MAS_VERBLVL_DEBUG, "anx: doing AUDIO_START playback ioctl" );
#endif
    err = ioctl( state->pdstate.play_fd, AUDIO_CONTROL, &actl );
    
    if( err < 0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: AUDIO_START playback ioctl returned %d, errno '%s'\n", err, strerror(errno) );
        return mas_error(MERR_INVALID);
    }


    return 0;
}

int32
pdanx_playback_stop( struct anx_state* state )
{
    int32 err;
    audio_control actl;
    
    actl.ioctl_request = AUDIO_STOP;
    actl.position = 0;
    actl.request_info = NULL;


#ifdef DEBUG
    masc_log_message( MAS_VERBLVL_DEBUG, "anx: doing AUDIO_STOP ioctl" );
#endif
    err = ioctl( state->pdstate.play_fd, AUDIO_CONTROL, &actl );
    
    if( err < 0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: AUDIO_STOP ioctl  returned %d, errno '%s'\n", err, strerror(errno) );
        return mas_error(MERR_INVALID);
    }
    return 0;
}

int32
pdanx_playback_pause( struct anx_state* state )
{
    /* well... I think pausing happens without this low level being
       involved... */
#if 0
    int32 err;
    audio_control actl;
    
    actl.ioctl_request = AUDIO_PAUSE;
    actl.position = 0;
    actl.request_info = NULL;


#ifdef DEBUG
    masc_log_message( MAS_VERBLVL_DEBUG, "anx: doing AUDIO_PAUSE ioctl" );
#endif
    err = ioctl( state->pdstate.play_fd, AUDIO_CONTROL, &actl );
    
    if( err < 0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: AUDIO_PAUSE ioctl  returned %d, errno '%s'\n", err, strerror(errno) );
        return mas_error(MERR_INVALID);
    }
    return 0;
#endif
}

int32
pdanx_playback_poll( struct anx_state* state, struct mas_data* data )
{
    struct aix_state*        pdst = &state->pdstate;
    int                      error;
    int32                    err;
    audio_buffer             ab;
    int device_buffer_used = 0;
    int report_xrun = 0;

#if 0
    {
        uint32 ts;
        static uint32 ts_ref = 0;
        
        masc_get_short_usec_ts( &ts );
        masc_log_message( 0, "pdanx_playback_poll:  %u", ts-ts_ref );
        ts_ref=ts;
    }
#endif

    if ( !state->play_buffer->filling )
    {
        /* check the buffer space used in the device */
        memset( &ab, 0, sizeof ab );
        err = ioctl( pdst->play_fd, AUDIO_BUFFER, &ab );
        if( err < 0 )
        {
            masc_log_message( MAS_VERBLVL_DEBUG, "anx: pdanx_playback_poll: AUDIO_BUFFER ioctl returned %d. errno is '%s'", err, strerror(errno) );
        }
        else
        {
            device_buffer_used = ab.write_buf_size;
        }
        
        /* trying to emulate OSS xrun detection */

        /* test for underruns */
        if ( (ab.flags & (AUDIO_UNDERRUN | AUDIO_FIFO_UNDERRUN)) != 0 )
            report_xrun = TRUE;
        
        /* if the device buffer has underrun, start filling it */
        if ( device_buffer_used == 0 )
            report_xrun = TRUE;
    }
    
    if ( state->rebuffer )
    {
        state->play_buffer->filling = TRUE;
        state->rebuffer = FALSE;
    }
    
    /** we've got to fill the buffer */
    if ( state->play_buffer->filling )
    {
        err = anx_buffer_append( state->play_buffer, data->segment, data->length );
        masc_log_message(MAS_VERBLVL_DEBUG, "anx: filling buffer - %%%d percent full", 100 * state->play_buffer->pos/state->play_buffer->fill_line );
        if ( err < 0) return err;
    }
    
    if (! state->play_buffer->filling ) /* catches changes from above */
    {
        /* Write anything in the buffer, all at once. */
        if (state->play_buffer->pos > 0)
        {
            /* set reference mark */
            state->mcref = state->mcnow;
            state->mtref = data->header.media_timestamp - ( state->play_buffer->pos - data->length ) / state->play_bpstc;
            state->valid_refmark = TRUE;
            
            /* write the buffer */
            if (( error = write (pdst->play_fd, state->play_buffer->contents, state->play_buffer->pos)) != state->play_buffer->pos )
            {
                if( error==-1 )
                {
                    masc_log_message( MAS_VERBLVL_DEBUG, "anx: playback write error '%s'", strerror(errno) );
                }
                else
                {
                    masc_log_message( MAS_VERBLVL_DEBUG, "anx: playback write only wrote %d out of %d bytes", error,  state->play_buffer->pos );
                }
                return mas_error(MERR_IO);
            }
            
            state->played_bytes += state->play_buffer->pos;
            anx_reset_buffer( state->play_buffer );
        }
        else
        {
            /* nothing in buffer - write this data's data */
            if (( error = write (pdst->play_fd, data->segment, data->length)) != data->length )
            {
                if( error==-1 )
                {
                    masc_log_message( MAS_VERBLVL_DEBUG, "anx: playback write error '%s'", strerror(errno) );
                }
                else
                {
                    masc_log_message( MAS_VERBLVL_DEBUG, "anx: playback write only wrote %d out of %d bytes", error,  state->play_buffer->pos );
                }
                return mas_error(MERR_IO);
            }
            
            state->played_bytes += data->length;
        }
    }
    
    /* we ran out of buffer */
    if ( report_xrun ) return mas_error(MERR_XRUN);
    
    return 0;
}

int32
pdanx_record_start( struct anx_state* state )
{
    struct aix_state*  pdst = &state->pdstate;
    int32 err;
    audio_control actl;
    audio_change  achg;
    float lvol;
    struct mixer_channel* mch = &(state->mch[2]);
    
    if( state->pdstate.rec_fd<0 )
    {
        return mas_error( MERR_INVALID );
    }

    /* If we were recording before, stopped, and have now restarted,
       there's gonna be a ton of junk in the buffer.  trash it! */
    if ( state->rec_bytes )
    {
        err = ioctl( state->pdstate.rec_fd, AUDIO_WAIT, NULL );
        if( err<0 )
        {
            masc_log_message( MAS_VERBLVL_DEBUG, "error %d draining kernel record audio buffers", err );
        }
    }

    
    unmute_mic( state );

    lvol = (mch->left + 636.0) / 636.0;
    
    initialize_audio_change( &achg );
    achg.input                = DEFAULT_REC_SOURCE;
    
    achg.volume  = (int)((float)0x7fffffff * lvol); /* this doesn't work */
    achg.balance = 0x3FFFFFFF;
    
    achg.monitor  = AUDIO_MONITOR_OFF;
    
    actl.ioctl_request = AUDIO_CHANGE;
    actl.request_info = &achg;
    
    err = ioctl( state->pdstate.rec_fd, AUDIO_CONTROL, &actl );
    if ( err<0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "rec AUDIO_CONTROL ioctl returned %d, errno '%s", err, strerror(errno) );
        return mas_error(MERR_INVALID);
    }

    
    actl.ioctl_request = AUDIO_START;
    actl.position = 0;
    actl.request_info = NULL;
    
#ifdef DEBUG
    masc_log_message( MAS_VERBLVL_DEBUG, "anx: doing record AUDIO_START ioctl" );
#endif
    err = ioctl( state->pdstate.rec_fd, AUDIO_CONTROL, &actl );
    
    if( err < 0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: AUDIO_START record ioctl returned %d, errno '%s'\n", err, strerror(errno) );
        return mas_error(MERR_INVALID);
    }
    
    return 0;
}

int32
pdanx_record_stop( struct anx_state* state )
{
    int32 err;
    audio_control actl;
    
    actl.ioctl_request = AUDIO_STOP;
    actl.position = 0;
    actl.request_info = NULL;


    if( state->pdstate.rec_fd<0 )
    {
        return mas_error( MERR_INVALID );
    }

#ifdef DEBUG
    masc_log_message( MAS_VERBLVL_DEBUG, "anx: doing record AUDIO_STOP ioctl" );
#endif
    err = ioctl( state->pdstate.rec_fd, AUDIO_CONTROL, &actl );
    
    if( err < 0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: record AUDIO_STOP ioctl returned %d, errno '%s'\n", err, strerror(errno) );
        return mas_error(MERR_INVALID);
    }

    return 0;
}

int32
pdanx_record_pause( struct anx_state* state )
{
    return 0;
}

int32
pdanx_record_poll( struct anx_state* state, struct mas_data** data_ptr )
{
    struct aix_state* pdst = &state->pdstate;
    struct mas_data* data;
    int amt;
    
    *data_ptr = 0;
#if 0    
    {
        uint32 ts;
        static uint32 ts_ref = 0;
        
        masc_get_short_usec_ts( &ts );
        masc_log_message( 0, "pdanx_record_poll:  %u", ts-ts_ref );
        ts_ref=ts;
    }
#endif
    data = MAS_NEW( data );
    masc_setup_data( data, MAS_ANX_SEGLEN );

    amt = read (pdst->rec_fd, data->segment, MAS_ANX_SEGLEN);
    
    if ( amt != MAS_ANX_SEGLEN )
    {
/*         masc_log_message( MAS_VERBLVL_DEBUG, "anx: pdanx_record_poll: not enough data (wanted %d, got %d)", MAS_ANX_SEGLEN, amt ); */
        if( amt<=0 )
        {
            masc_strike_data( data );
            masc_rtfree( data );
            return mas_error(MERR_IO) | mas_make_serror( errno );
        }
    }
    
    *data_ptr = data;
    return 0;
}

int32
pdanx_show_state( struct anx_state* state )
{
    /* TODO */
#if 0
    u_int32                  holder;
    audio_buf_info           playinfo;

    masc_log_message(0, "*-- platform dependent anx state -------------------------------\n");

    ioctl(state->pdstate.fd, SNDCTL_DSP_GETBLKSIZE, &holder);
    masc_log_message(0, "fragment size: %d", holder);

    memset( &playinfo, 0, sizeof playinfo );
    ioctl(state->pdstate.fd, SNDCTL_DSP_GETOSPACE, &playinfo);

    masc_log_message(0, "    fragments: %d", playinfo.fragments);
    masc_log_message(0, "   fragstotal: %d", playinfo.fragstotal);
    masc_log_message(0, "     fragsize: %d", playinfo.fragsize);
    masc_log_message(0, "        bytes: %d", playinfo.bytes);

    masc_log_message(0, "for recording");
    ioctl(state->pdstate.fd, SNDCTL_DSP_GETISPACE, &playinfo);

    masc_log_message(0, "    fragments: %d", playinfo.fragments);
    masc_log_message(0, "   fragstotal: %d", playinfo.fragstotal);
    masc_log_message(0, "     fragsize: %d", playinfo.fragsize);
    masc_log_message(0, "        bytes: %d", playinfo.bytes);
#endif
    return 0;
}


int32
pdanx_get( struct anx_state* state, char* key, struct mas_package* arg, struct mas_package* r_package )
{
    struct aix_state*  pdst = &state->pdstate;
    /* list of nuggets.  preserve the terminator */
    static char* nuggets[] =
        { "list", "gain_db", "gain_linear", "channels", "recsource",
          "mc_clocks", "buftime_ms", "mc_clkid", "res_state", "res_state_ito_s", "aix_audio_device", "aix_internal_speaker", "" };
    
    int remain;
    int i, n=0;
    
    /* count the defined nuggets */
    while ( *nuggets[n] != 0 ) n++;

    i = masc_get_string_index(key, nuggets, n);

    switch(i)
    {
        /* TODO */
        
    case 0: /*list*/
        masc_push_strings( r_package, nuggets, n );
        break;

    case 10: /*aix_audio_device*/
        masc_pushk_string( r_package, "aix_audio_device", state->pdstate.devfname );
        break;

    case 11: /*aix_internal_speaker*/
        masc_pushk_int8( r_package, "aix_internal_speaker", state->pdstate.internal_speaker_state );
        break; 

    default:
        return mas_error(MERR_INVALID);
        break;
    }

    return 0;
}

int32
pdanx_set( struct anx_state* state, char* key, struct mas_package* arg )
{
    int32 err;
    /* list of nuggets.  preserve the terminator */
    static char* nuggets[] = 
        { "gain_db", "gain_linear", "recsource", "mc_clock_addx", "buftime_ms", "res_state", "res_state_ito_s", "aix_audio_device", "aix_internal_speaker", "" };
    int i, n=0;
    
    while ( *nuggets[n] != 0 ) n++;
    
    i = masc_get_string_index(key, nuggets, n);
    
    switch(i)
    {
    case 7: /*aix_audio_device*/
    {
        char *name;
        masc_pullk_string( arg, "aix_audio_device", &name, FALSE );
        /* TODO: maybe not trust this input */
        strncpy( state->pdstate.devfname, name, 128 );
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: new device name is %s", state->pdstate.devfname );

        if( state->pdstate.play_configured )
            pdanx_close_resource( state );

        pdanx_open_resource( state );
        
        if( state->source_active )
        {
            pdanx_configure_resource( state, state->audio_source );
            pdanx_record_start( state ); /* for the AUDIO_START ioctl */
        }
        if( state->sink_active )
        {
            pdanx_configure_resource( state, state->audio_sink );
            pdanx_playback_start( state ); /* for the AUDIO_START ioctl */
        }
        break;
    }
    case 8: /*aix_internal_speaker*/
    {
        int8 bogus;
        audio_change achg;
        audio_control actl;
        
        masc_pull_int8( arg, &bogus );
        initialize_audio_change( &achg );
                
        achg.volume  = 0x7fffffff;
        achg.balance = 0x3FFFFFFF;

        state->pdstate.internal_speaker_state = bogus;
        if( state->pdstate.internal_speaker_state ) 
        {
            achg.output  = AUDIO_INTERNAL_AUDIO_ON;
        }
        else
        {
            achg.output  = AUDIO_INTERNAL_AUDIO_OFF;
        }
            
            
        actl.ioctl_request = AUDIO_CHANGE;
        actl.request_info  = &achg;
        
        err = ioctl( state->pdstate.play_fd, AUDIO_CONTROL, &actl );
        if ( err<0 ) 
        {
            masc_log_message( MAS_VERBLVL_DEBUG, "anx: ioctl changing audio interface returned %d, errno==%d (%s)", err, errno, strerror(errno) );
            return EXIT_FAILURE;
        }
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: aix internal speaker turned %s.", (state->pdstate.internal_speaker_state)?"ON":"OFF" );
        break;
    }
    default:
        break;
    }
    
    return 0;
}


int32
pdanx_set_mixer_volume(struct anx_state* state, int ch_id )
{
    int32 err;
    int left, right;
    struct mixer_channel* mch = &(state->mch[ch_id]); /* convenience */
    audio_channels_list acl;
    audio_channel_setting acset[2*ALLOCATED_MIX_CH];
    int sns_ch = state->pdstate.aix_mch_map[ch_id];
    int i, j, ach;
    float lvol, rvol, bal;    
    audio_set_gain asg;
    
    audio_control    actl;
    audio_change     achg;
    
    audio_channel_status acs;

    /* if the file descriptors are closed we can't set the
       volume. TODO: we may want to cache the desired values and set
       them when the device becomes active again. For now, just ignore
       the request silently. */  
    if( !(state->pdstate.play_configured || state->pdstate.play_configured) )
        return 0;

    if( aix_channel(sns_ch)==MASTER_VOLUME_CHANNEL )
    {
        /* adapter needs scale that goes up to 160 ?!?  */
        left = 160*(mch->left+636)/636;
        
        err = ioctl( state->pdstate.mstvol_fd, AUDIO_MASTER_VOLUME, &left );
        if ( err<0 ) 
        {
            masc_log_message( MAS_VERBLVL_DEBUG, "anx: AUDIO_MASTER_VOLUME setting ioctl returned %d. errno is '%s'", err, strerror(errno) );
            return mas_error(MERR_INVALID);
        }
#if 0
#ifdef DEBUG
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: pdanx_set_mixer_volume: master volume is %d", left );
#endif        
#endif
    }
    else
    {
        memset( &actl, 0, sizeof actl );
        initialize_audio_change( &achg );
        
        if( is_input_channel(sns_ch) )
        {
            achg.input = aix_channel(sns_ch);
        }
        else
        {
            achg.output = aix_channel(sns_ch);   
        }

        lvol = (mch->left  + 636.0) / 636.0;
        rvol = (mch->right + 636.0) / 636.0;

        if( mch->is_stereo )
        {
            /* heh, I wish I could see what the adapter does with
               these values. Hoping that it's just linear like this.
               (this also corresponds with the weird way masmix assigns balance) */
            if( (lvol==rvol) && (rvol==0.0) )
            {
                achg.volume  = 0;
                achg.balance = 0x3FFFFFFF;
            }
            else
            {
                if( lvol>rvol )
                {
                    bal = rvol / lvol - 1.0;
                    achg.volume  = (int)( (float)0x7fffffff * lvol );
                }
                else
                {
                    bal = 1.0 - lvol / rvol;
                    achg.volume  = (int)( (float)0x7fffffff * rvol );
                }
                achg.balance = 0x3FFFFFFF * (bal+1);
            }
        }
        else
        {
            achg.volume  = (float)0x7fffffff * lvol;
            achg.balance = AUDIO_IGNORE;
        }

        if( is_input_channel(sns_ch) )
        {
            asg.left_gain=100*lvol;
            asg.right_gain=100*lvol;
#if 0
            /* I found by experiment that 0..100 is the valid range for
            this. However, at 0 we still get considerable signal from
            the mic. What gives? */
            masc_log_message( MAS_VERBLVL_DEBUG, "AUDIO_SET_GAIN setting %d, %d", asg.left_gain, asg.right_gain );
#endif
            err = ioctl( is_input_channel(sns_ch) ? state->pdstate.rec_fd : state->pdstate.play_fd, AUDIO_SET_GAIN, &asg );
            if ( err<0 )
            {
                masc_log_message( MAS_VERBLVL_DEBUG, "AUDIO_SET_GAIN ioctl returned %d, errno '%s'\n", err, strerror(errno) );
                return mas_error(MERR_INVALID);
            }
        }
        else
        {
            actl.ioctl_request = AUDIO_CHANGE;
            actl.request_info = &achg;
            
            err = ioctl( is_input_channel(sns_ch) ? state->pdstate.rec_fd : state->pdstate.play_fd, AUDIO_CONTROL, &actl );

            if ( err<0 )
            {
                masc_log_message( MAS_VERBLVL_DEBUG, "ioctl changing audio interface returned %d, errno==%d '%s'", err, errno, strerror(errno) );
                return mas_error(MERR_INVALID);
            }
        }
    }
    return 0;
}


int32
pdanx_get_mixer_volume(struct anx_state* state, int ch_id )
{
    int left, right;
    struct mixer_channel* mch = &(state->mch[ch_id]); /* convenience */
    int32 err;
    int volume;
    audio_channel_status acs;
    audio_status         as;
    int mapped_ch =  state->pdstate.aix_mch_map[ch_id];
    float vol, bal;


    /* if the file descriptors are closed we can't query the
       volume. Just return without error */  
    if( !(state->pdstate.play_configured || state->pdstate.play_configured) )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: pdanx_get_mixer_volume: can't query volume (yet?) because we're not configured" );
        return 0;
    }
    
    
    if( aix_channel(mapped_ch)==MASTER_VOLUME_CHANNEL )
    {
        /* Betting that AUDIO_HOST_OUTPUT is always there */
        acs.channel_count    = 1;
        acs.channel_type[0]  = AUDIO_HOST_OUTPUT;
        acs.channel[0]       = AUDIO_LEFT_CHANNEL;
        
        err = ioctl( state->pdstate.mix_fd, AUDIO_CHANNEL_STATUS, &acs );
        if( err < 0 )
        {
            masc_log_message( MAS_VERBLVL_ERROR, "anx: [error] AUDIO_CHANNEL_STATUS ioctl returned %d, errno is '%s'\n", err, strerror(errno) );
            return mas_error( MERR_INVALID );
        }


        mch->left = (636*(acs.master_volume-160))/160;
        mch->right = mch->left; /* there seems to be no balance control on
                                 * the master volume on AIX */
#ifdef DEBUG
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: pdanx_get_mixer_volume: master volume is %d", acs.master_volume );
#endif        
        return 0;
    }
    else
    {
        memset( &as, 0, sizeof as );

        if( is_input_channel(mapped_ch) )
            err = ioctl( state->pdstate.rec_fd, AUDIO_STATUS, &as );
        else
            err = ioctl( state->pdstate.play_fd, AUDIO_STATUS, &as );
        
        if( err < 0 )
        {
            masc_log_message( MAS_VERBLVL_DEBUG, "anx: AUDIO_STATUS ioctl returned %d. errno %d '%s'\n", err, errno, strerror(errno) );
            return mas_error(MERR_INVALID);
        }
#ifdef DEBUG
        print_status( &as );
#endif
        vol = as.change.volume / (float)0x7fffffff;
        bal = (as.change.balance - (float)0x3fffffff) / (float)0x3fffffff; /* goes from -1 to 1 */
        
        /* TODO: balance */
        mch->left  = 636*(vol-1.0);
        mch->right = 636*(vol-1.0);
        

#ifdef DEBUG
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: pdanx_get_mixer_volume: aix channel %s's MAS volume is (%d, %d)", is_input_channel(mapped_ch)?change_input_symbol(aix_channel(mapped_ch)) : change_output_symbol(aix_channel(mapped_ch)), mch->left, mch->right );
#endif  
        return 0;
    }
}


int32
pdanx_set_recording_source(struct anx_state* state, int ch_id )
{
    audio_control actl;
    audio_change  achg;
    int32         err;
    
    if( state->pdstate.rec_fd<0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: pdanx_set_recording_source(): rec_fd not open?" );
        return mas_error( MERR_INVALID );
    }
    
    initialize_audio_change( &achg );
    achg.input  = state->pdstate.rec_source = state->pdstate.aix_mch_map[ ch_id ];
    
    achg.volume  = 0x7fffffff / 2; 
    
    achg.balance = 0x3FFFFFFF;
    
    actl.ioctl_request = AUDIO_CHANGE;
    actl.request_info = &achg;
    
    err = ioctl( state->pdstate.rec_fd, AUDIO_CONTROL, &actl );
    if ( err<0 ) 
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "rec AUDIO_CONTROL ioctl returned %d, errno '%s", err, strerror(errno) );
        return mas_error(MERR_INVALID);
    }

    return 0;
}

int32
pdanx_get_recording_source(struct anx_state* state )
{
    return state->pdstate.rec_source;
}

/*
 * The sample counter counts the number of N-channel samples that have
 * been successfully played by the interface. It will not increment if
 * nothing is playing, and it will not increment during buffer
 * underruns.  Its value is held in scnt->val in the master clock
 * value structure scnt.
 *
 * It is up to you to register a "veto" of the measurement if any
 * condition occurs that you believe could affect it.  In the case of
 * these OSS drivers, if the interface isn't playing or there are
 * buffer underruns, a veto is registered.  Additionally, the outer
 * layer of the anx device will veto the measurement if the sample
 * counter wraps.
 *
 * This function returns zero (0) if an error occurs, otherwise it
 * returns the sample counter's value.  This function must NOT return
 * a valid count when an error has occurred.  For the OSS drivers, we
 * check for playback underruns immediately after retrieving the
 * sample counter.
 *
 * The counter only counts at the sampling frequency, and is
 * independent of the number of channels or resolution of the samples:
 * e.g. for 44.1 kHz, 16-bit, stereo, the counter will read 44100
 * after one second, 88200 after 2 seconds, and so on.
 *
 */

uint32
pdanx_get_sample_count( struct anx_state* state, struct mas_mc_clkval* scnt )
{
    int device_buffer_used = 0;
    unsigned int bytes;
    audio_buffer ab;
    int32 err;
    
    if ( scnt == 0 ) return 0;


    /* check the buffer space used in the device */
    memset( &ab, 0, sizeof ab );
    if( state->pdstate.play_configured )
    {
        err = ioctl( state->pdstate.play_fd, AUDIO_BUFFER, &ab );
        if( err < 0 )
        {
            masc_log_message( MAS_VERBLVL_DEBUG, "anx: get_sample_count: AUDIO_BUFFER ioctl returned %d. errno is '%s'\n", err, strerror(errno) );
            goto fail;
        }
        else
        {
            device_buffer_used = ab.write_buf_size;
        }
    }
    else
    {
        goto fail;
    }
    

    /* yargh. I think we don't get anything but this millisecond
       value. We'd really like to have bytes :( */
    bytes = ab.position * state->play_sample_rate * state->play_bpstc / 1000;
    
    /* Fail if the device buffer has underrun! */
    if ( device_buffer_used == 0 )
        goto fail;

    /* Otherwise, counter is valid; set the MC clock value to the
     * sample count. */
    scnt->val = bytes / state->play_bpstc;

    return scnt->val;

 fail:
    scnt->veto = TRUE;

    return 0;
    
}

/********************************************************************
  LOCAL FUNCTIONS
********************************************************************/

static int32
get_device_caps( struct anx_state* state )
{
    struct aix_state*  pdst = &state->pdstate;
    int sm = 0;
    int i;
    int32 err;
    audio_get_channels agc;
    int ahchannel = -1;
    
    /* full duplex? */
    if (  state->is_full_duplex==0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "device has half-duplex capability");
    }
    else
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "device has full-duplex capability");
    }

    /* I think we're not sample or fragment accurate. */
    state->is_sample_accurate  = 0;
    pdst->is_fragment_accurate = 0;
    masc_log_message( MAS_VERBLVL_DEBUG, "device may not be sample   accurate" );
    masc_log_message( MAS_VERBLVL_DEBUG, "device may not be fragment accurate" );
    

    /* could we MMAP the DMA buffer?  (not used) */
    pdst->is_mmap_able = 0;
    masc_log_message( MAS_VERBLVL_DEBUG, "won't use mmap-able buffers (not implemented)");

    
    /** MIXER ************/
    /* (make sure to init the is_stereo bit before calling pdanx_get_mixer_volume) */
        
/*     masc_log_message( MAS_VERBLVL_DEBUG, "mixer supports main"); */
/*     state->main_mix_ch = aix_add_mix_channel( state, "main mix", -1, 0 ); /\* !! the '0' is an index into the sensible_channel_map *\/ */
/*     /\* master volume seems to be mono on AIX: *\/ */
/*     state->mch[state->main_mix_ch].is_stereo = FALSE; */
/*     pdanx_get_mixer_volume( state, state->main_mix_ch ); */
 

/*     /\* we're not using the get/set_channels method any more, so the following */
/*        is purely informational *\/ */
/* #ifdef DEBUG */
/*     /\* get information about all supported channels and what they can do *\/ */
/*     err = ioctl( pdst->mix_fd, AUDIO_GET_CHANNELS, &agc ); */
/*     if( err<0 ) */
/*     { */
/*         masc_log_message( MAS_VERBLVL_ERROR, "anx: [error] doing AUDIO_GET_CHANNELS ioctl; errno is '%s'", strerror(errno) ); */
/*     } */
/*     print_get_channels( &agc ); */
/* #endif */

    /* this is hardcoded now. We used to detect it when we were using
       the iomix AUDIO_GET_CHALLELS stuff... see older revisions of
       this file */
    masc_log_message( MAS_VERBLVL_DEBUG, "mixer supports dac");
    state->dac_ch = aix_add_mix_channel( state, "dac", state->audio_sink, 1 );
    state->mch[state->dac_ch].is_stereo = TRUE;
/* do this later, when we're configured     pdanx_get_mixer_volume( state, state->dac_ch ); */

    masc_log_message( MAS_VERBLVL_DEBUG, "mixer supports mic");
    state->mic_ch = aix_add_mix_channel( state, "mic", state->audio_sink, 2 );
    state->mch[state->mic_ch].is_stereo = FALSE;
/*  do this later, when we're configured   pdanx_get_mixer_volume( state, state->mic_ch ); */
    
    return 0;
}


int32
aix_set_fragment_stuff( struct anx_state* state )
{
    struct aix_state* pdst = &state->pdstate; 
    audio_init_buf_req abr;
    int fragsize_pow2;
    int bail, fail;
    int32 err;
    
    /*** COMPUTE FRAGMENT SIZE AND NUMBER OF FRAGMENTS ***************/

    /* I'm assuming here that the OSS term "fragment" means the same
        as the AIX term "block" (and don't really know what either of
        them actually are :( my working assumption is that it's
        somehow the smallest unit that the driver internally pushes around? ) */


    /* TODO: to get things running, I'm only trying one set of values here.
       At a later point, come back and tweak. Also, maybe these values
       are too low to guarantee smooth operation */
    
    abr.buffer_size      = 2048;  /* basically, all of these are too
                                   * low. I'm hoping the driver will
                                   * respond with its smallest
                                   * possible values */
    abr.play_block_count = 10;
    abr.play_lower_limit = 10;
    abr.cap_upper_limit  = 10;
    abr.byte_count       = 10;
    abr.return_code      = AUDIO_IGNORE;

    
    err = ioctl( pdst->play_fd, AUDIO_MODIFY_LIMITS, &abr );
    if( err < 0 )
    {
        masc_log_message( MAS_VERBLVL_ERROR, "anx: [error] AUDIO_MODIFY_LIMITS ioctl returned %d, errno is '%s', return code %d\n", err, strerror(errno), abr.return_code );
        return mas_error( MERR_INVALID );
    }

    pdst->fragsize  = 128; /* see AUDIO_INIT ioctl! */

    if( abr.buffer_size % pdst->fragsize )
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: buffer size %d is not an integer multiple of fragment size %d", abr.buffer_size, pdst->fragsize );
    
    pdst->fragments = abr.buffer_size / pdst->fragsize; /* will this work? */

#ifdef DEBUG
    masc_log_message( MAS_VERBLVL_DEBUG, "--[ RESULTS OF PLAY MODIFY_LIMITS ]-" );
    masc_log_message( MAS_VERBLVL_DEBUG, "buffer_size          %d", abr.buffer_size );
    masc_log_message( MAS_VERBLVL_DEBUG, "play_block_count     %d", abr.play_block_count );
    masc_log_message( MAS_VERBLVL_DEBUG, "play_lower_limit     %d", abr.play_lower_limit );
    masc_log_message( MAS_VERBLVL_DEBUG, "cap_upper_limit      %d", abr.cap_upper_limit );
    masc_log_message( MAS_VERBLVL_DEBUG, "byte_count           %d", abr.byte_count );
    masc_log_message( MAS_VERBLVL_DEBUG, "return_code          %d", abr.return_code );
    masc_log_message( MAS_VERBLVL_DEBUG, "------------------------------------" );
#endif

    return 0;
}

int32
aix_set_rec_fragment_stuff( struct anx_state* state )
{
    struct aix_state* pdst = &state->pdstate; 
    audio_init_buf_req abr;
    int fragsize_pow2;
    int bail, fail;
    int32 err;

    if( state->pdstate.rec_fd<0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "anx: pdanx_set_recording_source(): rec_fd not open?" );
        return mas_error( MERR_INVALID );
    }

    abr.buffer_size      = 2048;  
    abr.play_block_count = 10;
    abr.play_lower_limit = 10;
    abr.cap_upper_limit  = 10;
    abr.byte_count       = 10;
    abr.return_code      = AUDIO_IGNORE;
    
    err = ioctl( pdst->rec_fd, AUDIO_MODIFY_LIMITS, &abr );
    if( err < 0 )
    {
        masc_log_message( MAS_VERBLVL_ERROR, "anx: [error] AUDIO_MODIFY_LIMITS ioctl returned %d, errno is '%s', return code %d\n", err, strerror(errno), abr.return_code );
        return mas_error( MERR_INVALID );
    }
    
#ifdef DEBUG
    masc_log_message( MAS_VERBLVL_DEBUG, "--[ RESULTS OF REC MODIFY_LIMITS ]--" );
    masc_log_message( MAS_VERBLVL_DEBUG, "buffer_size          %d", abr.buffer_size );
    masc_log_message( MAS_VERBLVL_DEBUG, "play_block_count     %d", abr.play_block_count );
    masc_log_message( MAS_VERBLVL_DEBUG, "play_lower_limit     %d", abr.play_lower_limit );
    masc_log_message( MAS_VERBLVL_DEBUG, "cap_upper_limit      %d", abr.cap_upper_limit );
    masc_log_message( MAS_VERBLVL_DEBUG, "byte_count           %d", abr.byte_count );
    masc_log_message( MAS_VERBLVL_DEBUG, "return_code          %d", abr.return_code );
    masc_log_message( MAS_VERBLVL_DEBUG, "------------------------------------" );
#endif
    
    return 0;
}


/* sets up a new mixer channel.  returns -1 on error, or the ID of the
   new channel if successful. */
int32
aix_add_mix_channel(struct anx_state* state, char* name, int32 portnum, int aix_ch_num)
{
    int32 i;

    /* use platform-independent mixer channel add */
    i = add_mix_channel(state->mch, name, portnum, ALLOCATED_MIX_CH );
    if ( i < 0 ) return i;
    
    state->pdstate.aix_mch_map[i] = aix_ch_num;
    
    return i;
}



int
aix_open_audio_device_playrec( struct anx_state* state )
{
    int fd;
    char *fname;
    
    if( (state->pdstate.devfname==NULL) || (*state->pdstate.devfname==0) )
        return mas_error( MERR_NULLPTR );
    
        
    fname = malloc( strlen(state->pdstate.devfname) + 3 );
    if( fname==NULL )
        return mas_error( MERR_MEMORY );

    strcpy( fname, state->pdstate.devfname );
    strcat( fname, "/1" );
    
    if ( (fd = open(fname, O_WRONLY|O_NDELAY)) < 0 )
    {
        masc_log_message( MAS_VERBLVL_ERROR, "anx: [error] Couldn't open audio device for playback.");
        masc_log_message( MAS_VERBLVL_ERROR, "anx: [error] errno is '%s'", strerror(errno) );

        free( fname );
        return mas_error(MERR_FILE_CANNOT_OPEN);
    }

    state->pdstate.play_fd = fd;
    state->is_full_duplex = TRUE;
    
    strcpy( fname, state->pdstate.devfname );
    strcat( fname, "/2" );

    if ( (fd = open(fname, O_RDONLY|O_NDELAY)) < 0 )
    {
        masc_log_message( MAS_VERBLVL_INFO, "anx: [info] Couldn't open audio device for recording.");
        masc_log_message( MAS_VERBLVL_INFO, "anx: [info] errno is '%s'", strerror(errno) );
        masc_log_message( MAS_VERBLVL_INFO, "anx: [info] audio device open for playback only.");        
        state->is_full_duplex = FALSE;
    }
    else
    {
        state->pdstate.rec_fd = fd;
        masc_log_message( MAS_VERBLVL_INFO, "anx: [info] audio device open for simultaneous recording and playback.");
    }
        
    free( fname );
    return  state->pdstate.play_fd;
}

int
aix_open_audio_device_mix( struct anx_state* state )
{
    int fd;
    char *fname;
    
    if( (state->pdstate.devfname==NULL) || (*state->pdstate.devfname==0) )
        return mas_error( MERR_NULLPTR );
    
        
    fname = malloc( strlen(state->pdstate.devfname) + 7 );
    if( fname==NULL )
        return mas_error( MERR_MEMORY );

    strcpy( fname, state->pdstate.devfname );
    strcat( fname, "/iomix" );
    
    if ( (fd = open(fname, O_RDWR|O_NDELAY)) < 0 )
    {
        masc_log_message( MAS_VERBLVL_ERROR, "anx: [error] Couldn't open mixer channel of audio device.");
        masc_log_message( MAS_VERBLVL_ERROR, "anx: [error] errno is '%s'", strerror(errno) );

        free( fname );
        return mas_error(MERR_FILE_CANNOT_OPEN);
    }
    
    state->pdstate.mix_fd = fd;
    
    free( fname );
    return  state->pdstate.mix_fd;
}


int
aix_open_audio_device_mstvol( struct anx_state* state )
{
    int fd;
    char *fname;
    
    if( (state->pdstate.devfname==NULL) || (*state->pdstate.devfname==0) )
        return mas_error( MERR_NULLPTR );
    
        
    fname = malloc( strlen(state->pdstate.devfname) + 7 );
    if( fname==NULL )
        return mas_error( MERR_MEMORY );

    strcpy( fname, state->pdstate.devfname );
    strcat( fname, "/mstvol" );
    
    if ( (fd = open(fname, O_WRONLY|O_NDELAY)) < 0 )
    {
        masc_log_message( MAS_VERBLVL_ERROR, "anx: [error] Couldn't open master volume channel of audio device.");
        masc_log_message( MAS_VERBLVL_ERROR, "anx: [error] errno is '%s'", strerror(errno) );

        free( fname );
        return mas_error(MERR_FILE_CANNOT_OPEN);
    }
    
    state->pdstate.mstvol_fd = fd;
    
    free( fname );
    return  state->pdstate.mstvol_fd;
}


/* the Power PC AIX adapter uses several different gain tables,
 * i.e. you can vary levels only in discrete decibel steps of certain
 * sizes within certain limits; the sizes and limits vary between
 * channel types. See audiodd2.html for definitions.
 *
 * Therefore, THE FOLLOWING FUNCTION WILL QUANTIZE (as in truncate) a
 * db value you give them to the next supported value, and return the
 * tabulated integer value for it.
 */
int
tabulated_gain( int aix_channel, float db )
{
    /* db values are between 0 and -60 */
    switch( aix_channel )
    {
    case AUDIO_MONO_OUT:
    case AUDIO_HOST_OUTPUT:

        /* six bit gain, 0 to -94db in 1.5db steps */
        if( db < -94.5 )
            db = -94.5;
        
        if( db > 0.0 )
            db = 0.0;
        
        return -db / 1.5;

        
    case AUDIO_HIGH_GAIN_MIC:
    case AUDIO_LOW_GAIN_MIC:
    case AUDIO_MONO_IN:

        /* four bit gain, 0 to -45db in 3db steps */ 
        if( db < -45 )
            db = -45;
        
        if( db > 0 )
            db = 0;
        
        return -db / 3;


    case AUDIO_LINE1_IN:
    case AUDIO_AUX1_IN:
    case AUDIO_AUX2_IN:
        
        /* five bit, 12..-24.5db in 1.5db steps */
        if( db > 12 )
            db = 12;
        
        if( db < -34.5 )
            db = -34.5;
        
        return (12-db)/1.5 ;

        
    case AUDIO_HOST_INPUT:

        db = 22.5*(db+60)/60; /* ? */
        
        /* four bit, 0 to 25db in 1.5db steps */
        if( db > 22.5 )
            db = 22.5;
        
        if( db < 0.0 )
            db = 0.0;
        
        return db / 1.5;

    default:
        return 0;
    }
}


float
inverse_tabulated_gain( int aix_channel, int gain )
{
    switch( aix_channel )
    {
    case AUDIO_MONO_OUT:
    case AUDIO_HOST_OUTPUT:

        /* six bit gain, 0 to -94db in 1.5db steps */
        return -1.5*gain;
        
        
    case AUDIO_HIGH_GAIN_MIC:
    case AUDIO_LOW_GAIN_MIC:
    case AUDIO_MONO_IN:

        /* four bit gain, 0 to -45db in 3db steps */ 
        return -3*gain;


    case AUDIO_LINE1_IN:
    case AUDIO_AUX1_IN:
    case AUDIO_AUX2_IN:

        /* five bit, 12..-24.5db in 1.5db steps */
        
        return 12-1.5*gain;
        
        
    case AUDIO_HOST_INPUT:

        /* four bit, 0 to 25db in 1.5db steps */        
        return 1.5*gain;

    default:
        return 0;
    }
}

void
initialize_audio_change( audio_change *ac )
{
    audio_track_info ati;

    memset( &ati, 0, sizeof ati );

    ati.master_volume  = 100;
    ati.dither_percent =   0;
    
    ac->dev_info      = &ati;
    ac->input         = AUDIO_IGNORE;        
    ac->output        = AUDIO_IGNORE;       
    ac->monitor       = AUDIO_IGNORE;      
    ac->volume        = AUDIO_IGNORE;       
    ac->volume_delay  = 0;
    ac->balance       = AUDIO_IGNORE;      
    ac->balance_delay = 0;
    ac->treble        = 0;       
    ac->bass          = 0;         
    ac->pitch         = 0;        
}


int32
unmute_mic( struct anx_state* state )
{
    int32 err;
    audio_channels_list acl;
    audio_channel_setting acset[1];
    int ach;
    
    
    ach = AUDIO_HIGH_GAIN_MIC;
    acl.csptr = acset;

    acset[0].channel_type = ach;
    acset[0].channel      = AUDIO_LEFT_CHANNEL;
            
    acset[0].channel_gain = tabulated_gain( ach, 0 );
    acset[0].gain_delay   = 0;
    acset[0].mute         = 0;

    acl.size = 1;
        
    err = ioctl( state->pdstate.mix_fd, AUDIO_SET_CHANNELS, &acl );
    if( err < 0 )
    {
        masc_log_message( MAS_VERBLVL_DEBUG, "AUDIO_SET_CHANNELS ioctl returned %d, errno is '%s'", err, strerror(errno) );
        return mas_error( MERR_INVALID );
    }

    return 0;
}
